#ifndef RESCUE_COMMON_H
#define RESCUE_COMMON_H

#include <vector>
#include <list>

#define INT_32_SIZE 4
#define INT_16_SIZE 2

namespace Librescue {

  typedef int INT_32;
  typedef short INT_16;

  typedef INT_32 Id;
  typedef INT_32 Cursor;
  // For backward compatability
  typedef INT_32 S32;

  class RescueObject;
  class Command;
  
  typedef unsigned char	Byte;
  typedef std::vector<Byte> Bytes;
  typedef std::vector<RescueObject*> Objects;
  typedef std::vector<Command*> Commands;
  typedef std::vector<Id> Ids;

#ifdef __OLD_CONSTANTS__
  enum Header {
	HEADER_NULL = 0x00,
	// GIS
	GK_CONNECT_OK = 0x30,
	GK_CONNECT_ERROR = 0x31,
	KG_CONNECT = 0x70,
	KG_ACKNOWLEDGE = 0x71,
	KG_UPDATE = 0x63,
	// Simulators
	SK_CONNECT = 0x20,
	SK_ACKNOWLEDGE = 0x21,
	SK_UPDATE = 0x22,
	KS_CONNECT_OK = 0x60,
	KS_CONNECT_ERROR = 0x61,
	KS_COMMANDS = 0x62,
	KS_UPDATE = 0x63,
	// Viewer
	VK_CONNECT = 0x20,
	VK_ACKNOWLEDGE = 0x21,
	KV_CONNECT_OK = 0x60,
	KV_CONNECT_ERROR = 0x61,
	KV_UPDATE = 0x63,
	// Agents
	AK_CONNECT = 0x10,
	AK_ACKNOWLEDGE = 0x11,
	KA_CONNECT_OK = 0x50,
	KA_CONNECT_ERROR = 0x51,
	KA_SENSE = 0x52,
	KA_HEAR = 0x53,
	KA_HEAR_SAY = 0x54,
	KA_HEAR_TELL = 0x55,
	// Agent commands	
	AK_REST = 0x80,
	AK_MOVE = 0x81,
	AK_LOAD = 0x82,
	AK_UNLOAD = 0x83,
	AK_SAY = 0x84,
	AK_TELL = 0x85,
	AK_EXTINGUISH = 0x86,
	AK_STRETCH = 0x87,
	AK_RESCUE = 0x88,
	AK_CLEAR = 0x89
  };
#else
  enum Header {
	HEADER_NULL = 0x00,
	// GIS
	GK_CONNECT_OK = 0x10,
	GK_CONNECT_ERROR = 0x11,
	KG_CONNECT = 0x12,
	KG_ACKNOWLEDGE = 0x13,
	KG_UPDATE = 0x14,
	// Simulators
	SK_CONNECT = 0x20,
	SK_ACKNOWLEDGE = 0x21,
	SK_UPDATE = 0x22,
	KS_CONNECT_OK = 0x23,
	KS_CONNECT_ERROR = 0x24,
	KS_COMMANDS = 0x25,
	KS_UPDATE = 0x26,
	// Viewer
	VK_CONNECT = 0x30,
	VK_ACKNOWLEDGE = 0x31,
	KV_CONNECT_OK = 0x32,
	KV_CONNECT_ERROR = 0x33,
	KV_UPDATE = 0x34,
	// Agents
	AK_CONNECT = 0x41,
	AK_ACKNOWLEDGE = 0x42,
	KA_CONNECT_OK = 0x43,
	KA_CONNECT_ERROR = 0x44,
	KA_SENSE = 0x45,
	KA_HEAR = 0x46,
	KA_HEAR_SAY = 0x47,
	KA_HEAR_TELL = 0x48,
	// Agent commands	
	AK_REST = 0x80,
	AK_MOVE = 0x81,
	AK_LOAD = 0x82,
	AK_UNLOAD = 0x83,
	AK_SAY = 0x84,
	AK_TELL = 0x85,
	AK_EXTINGUISH = 0x86,
	AK_STRETCH = 0x87,
	AK_RESCUE = 0x88,
	AK_CLEAR = 0x89
  };
#endif

  enum TypeId {
	TYPE_NULL = 0,
	TYPE_WORLD = 208,
	TYPE_CIVILIAN = 232,
	TYPE_CAR = 236,
	TYPE_FIRE_BRIGADE = 233,
	TYPE_AMBULANCE_TEAM = 234,
	TYPE_POLICE_FORCE = 235,
	TYPE_BUILDING = 176,
	TYPE_REFUGE = 184,
	TYPE_FIRE_STATION = 185,
	TYPE_AMBULANCE_CENTER = 186,
	TYPE_POLICE_OFFICE = 187,
	TYPE_ROAD = 168,
	TYPE_RIVER = 169,
	TYPE_NODE = 200,
	TYPE_RIVER_NODE = 201
  };

  enum PropertyId {
	PROPERTY_MIN = 1,
	PROPERTY_NULL = 0,
	PROPERTY_START_TIME = 29,
	PROPERTY_LONGITUDE = 30,
	PROPERTY_LATITUDE = 31,
	PROPERTY_WIND_FORCE = 32,
	PROPERTY_WIND_DIRECTION = 33,
	PROPERTY_POSITION = 6,
	PROPERTY_POSITION_EXTRA = 7,
	PROPERTY_DIRECTION = 27,
	PROPERTY_POSITION_HISTORY = 207,
	PROPERTY_STAMINA = 9,
	PROPERTY_HP = 10,
	PROPERTY_DAMAGE = 11,
	PROPERTY_BURIEDNESS = 23,
	PROPERTY_WATER_QUANTITY = 25,
	PROPERTY_STRETCHED_LENGTH = 26,
	PROPERTY_X = 3,
	PROPERTY_Y = 4,
	PROPERTY_FLOORS = 14,
	PROPERTY_BUILDING_ATTRIBUTES = 15,
	PROPERTY_IGNITION = 48,
	PROPERTY_FIERYNESS = 16,
	PROPERTY_BROKENNESS = 17,
	PROPERTY_ENTRANCES = 235,
	PROPERTY_BUILDING_CODE = 50,
	PROPERTY_BUILDING_AREA_GROUND = 51,
	PROPERTY_BUILDING_AREA_TOTAL = 52,
	PROPERTY_BUILDING_APEXES = 131,
	PROPERTY_HEAD = 12,
	PROPERTY_TAIL = 13,
	PROPERTY_LENGTH = 24,
	PROPERTY_ROAD_KIND = 19,
	PROPERTY_CARS_PASS_TO_HEAD = 34,
	PROPERTY_CARS_PASS_TO_TAIL = 35,
	PROPERTY_HUMANS_PASS_TO_HEAD = 36,
	PROPERTY_HUMANS_PASS_TO_TAIL = 37,
	PROPERTY_WIDTH = 38,
	PROPERTY_BLOCK = 22,
	PROPERTY_REPAIR_COST = 39,
	PROPERTY_MEDIAN_STRIP = 40,
	PROPERTY_LINES_TO_HEAD = 41,
	PROPERTY_LINES_TO_TAIL = 42,
	PROPERTY_WIDTH_FOR_WALKERS = 43,
	PROPERTY_EDGES = 242,
	PROPERTY_SIGNAL = 44,
	PROPERTY_SHORTCUT_TO_TURN = 128,
	PROPERTY_POCKET_TO_TURN_ACROSS = 129,
	PROPERTY_SIGNAL_TIMING = 130,
	PROPERTY_MAX = 255
  };

  enum AgentType {
	AGENT_TYPE_CIVILIAN = 0x01,
	AGENT_TYPE_FIRE_BRIGADE = 0x02,
	AGENT_TYPE_FIRE_STATION = 0x04,
	AGENT_TYPE_AMBULANCE_TEAM = 0x08,
	AGENT_TYPE_AMBULANCE_CENTER = 0x10,
	AGENT_TYPE_POLICE_FORCE = 0x20,
	AGENT_TYPE_POLICE_OFFICE = 0x40
  };
}

#endif
